import typing
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


@dataclass
class ScrapperData:
    topic: str
    candidate: typing.Any
    tasks: typing.List['TaskData'] = field(default_factory=list)
    auth_session: typing.Any = None
    error_message: str = None
    error_extra: dict = None
    is_valid: bool = True


@dataclass
class TaskData:
    action: str
    action_data: 'ActionData' = None
    criteria: typing.Dict = field(default_factory=dict)


@dataclass
class ActionData:
    fetched_data: typing.Any = None
    transformed_data: typing.Any = None



import json
import time

from starling.config import CONFIG
from starling.exception import RetryTaskExitError, RetryTaskError, RetryTaskSkipAuthError
from starling.helper import retry_task, EnhancedJSONEncoder
from starling.types import ScrapperData, TaskData
from .blueprint_scrapper import BlueprintScrapper


class Scrapper:
    @staticmethod
    def _run_task(task: TaskData, scrapper_data: ScrapperData):
        package_name, classname = task.action.rsplit('.', 1)
        blueprint = getattr(__import__(package_name, fromlist=[package_name]), classname)(scrapper_data, task)
        blueprint.fetch()
        blueprint.transform()
        time.sleep(blueprint.interval())
        return blueprint.action_data

    @staticmethod
    @retry_task(times=CONFIG.get('RETRY_TASK_COUNT'))
    def _run(scrapper: BlueprintScrapper, is_auth=True):
        if is_auth:
            scrapper.authenticate()
        for task in [task for task in scrapper.data.tasks if task.action_data is None]:
            task.action_data = Scrapper._run_task(task, scrapper.data)

    @staticmethod
    def run(scrapper: BlueprintScrapper, json_output=True):
        try:
            Scrapper._run(scrapper)
            return json.dumps(scrapper.data, cls=EnhancedJSONEncoder) if json_output else scrapper.data
        except (RetryTaskExitError, RetryTaskError, RetryTaskSkipAuthError) as e:
            scrapper.data.is_valid = False
            scrapper.data.error_message = e.message
            scrapper.data.error_extra = e.extra
            return json.dumps(scrapper.data, cls=EnhancedJSONEncoder) if json_output else scrapper.data



from typing import Text

import requests
from bs4 import BeautifulSoup


def get_soap(url: Text, **kw) -> BeautifulSoup:
    # TODO Retry Logic required and add test case
    res = requests.get(url, headers={'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36'})
    return BeautifulSoup(res.content, "html.parser")


def get_selenium(url: Text, **kw) -> None:
    # TODO Implementation required
    pass



from starling.app import Scrapper
from starling.blueprint_action import BlueprintAction
from starling.blueprint_scrapper import BlueprintScrapper
from starling.config import CONFIG
from starling.exception import RetryTaskExitError, RetryTaskSkipAuthError
from starling.types import ScrapperData


class MyRetryTaskExitErrorScrapper(BlueprintScrapper):
    @property
    def actions(self):
        return [
            'tests.starling.test_custom_error.MyActionWithRetryTaskExitError1',
            'tests.starling.test_custom_error.MyActionWithRetryTaskExitError2'
        ]

    def authenticate(self):
        pass


class MyActionWithRetryTaskExitError1(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = self.action_data.fetched_data if self.action_data is None else 0 + 1

    def transform(self):
        raise RetryTaskExitError('ip blocked')


class MyActionWithRetryTaskExitError2(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = self.action_data.fetched_data if self.action_data is None else 0 + 1

    def transform(self):
        raise RetryTaskExitError('ip blocked')


class MyRetryTaskSkipAuthErrorScrapper(BlueprintScrapper):
    @property
    def actions(self):
        return [
            'tests.starling.test_custom_error.MyActionWithRetryTaskSkipAuthError1',
            'tests.starling.test_custom_error.MyActionWithRetryTaskSkipAuthError2'
        ]

    def authenticate(self):
        pass


retry_task_skip_auth_error_counter = 0


class MyActionWithRetryTaskSkipAuthError1(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = 1

    def transform(self):
        self.action_data.transformed_data = self.action_data.fetched_data


class MyActionWithRetryTaskSkipAuthError2(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = 2

    def transform(self):
        global retry_task_skip_auth_error_counter
        retry_task_skip_auth_error_counter += 1
        raise RetryTaskSkipAuthError('retry task but skip auth', extra={'id': '12345'})


def test_scrapper_with_retry_task_exit_error():
    res = Scrapper.run(MyRetryTaskExitErrorScrapper(ScrapperData('place', {'place_id': '123123'}))
                       , json_output=False)

    assert "ip blocked" == res.error_message
    assert res.tasks[0].action_data is None
    assert res.tasks[1].action_data is None
    assert res.is_valid is False


def test_scrapper_with_retry_task_skip_auth_error():
    res = Scrapper.run(MyRetryTaskSkipAuthErrorScrapper(ScrapperData('place', {'place_id': '123123'}))
                       , json_output=False)
    global retry_task_skip_auth_error_counter
    assert "retry task but skip auth" == res.error_message
    assert res.error_extra['id'] == '12345'
    assert res.tasks[0].action_data.fetched_data == 1
    assert res.tasks[1].action_data is None
    assert res.is_valid is False
    # TODO should search another better ways that check retry count properly
    assert retry_task_skip_auth_error_counter == CONFIG.get('RETRY_TASK_COUNT')



import dataclasses
import json
from typing import Callable

from starling.exception import RetryTaskExitError, RetryTaskSkipAuthError, RetryTaskError


def singleton(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


def retry_task(times: int) -> Callable:
    def decorator(fn: Callable) -> Callable:
        def fn_retry(*args, **kwargs):
            attempt = 0
            is_auth = True
            while attempt < times - 1:
                try:
                    return fn(*args, **kwargs, is_auth=is_auth)
                except RetryTaskExitError as e:
                    raise e
                except RetryTaskSkipAuthError as e:
                    is_auth = False
                    attempt += 1
                except RetryTaskError as e:
                    attempt += 1
            return fn(*args, **kwargs, is_auth=is_auth)

        return fn_retry

    return decorator


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



import abc

from starling.types import ScrapperData, TaskData


class BlueprintScrapper(abc.ABC):
    def __init__(self, data: ScrapperData, deny_actions=None):
        if deny_actions is None:
            deny_actions = []
        self.data = data
        self.deny_actions = deny_actions

        for action in [item for item in self.actions if item not in self.deny_actions]:
            self.data.tasks += self.tasks(action)

    def tasks(self, action):
        return [TaskData(action=action)]

    @property
    @abc.abstractmethod
    def actions(self):
        return []

    @abc.abstractmethod
    def authenticate(self):
        pass



import json

import pytest

from starling.app import Scrapper
from starling.blueprint_action import BlueprintAction
from starling.blueprint_scrapper import BlueprintScrapper
from starling.types import ScrapperData, TaskData


class MyScrapper(BlueprintScrapper):
    def tasks(self, action):
        if action == 'tests.starling.test_app.MyAction1':
            return [
                TaskData(action=action, criteria={'from': '20191001', 'to': '20191031'}),
                TaskData(action=action, criteria={'from': '20191101', 'to': '20191130'})
            ]
        else:
            return [TaskData(action=action)]

    @property
    def actions(self):
        return [
            'tests.starling.test_app.MyAction1',
            'tests.starling.test_app.MyAction2',
            'tests.starling.test_app.MyAction3'
        ]

    def authenticate(self):
        pass


class MyAttributeErrorScrapper(BlueprintScrapper):
    @property
    def actions(self):
        return [
            'tests.starling.test_app.MyMyAction1'
        ]

    def authenticate(self):
        pass


class MyAction1(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = self.scrapper_data.candidate

    def transform(self):
        self.action_data.transformed_data = self.action_data.fetched_data

    def interval(self):
        return 0


class MyAction2(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = self.scrapper_data.candidate

    def transform(self):
        self.action_data.transformed_data = self.action_data.fetched_data


class MyAction3(BlueprintAction):
    def fetch(self):
        self.action_data.fetched_data = self.scrapper_data.candidate

    def transform(self):
        self.action_data.transformed_data = self.action_data.fetched_data


def test_scrapper():
    res = Scrapper.run(MyScrapper(ScrapperData('place', {'place_id': '123123'})), json_output=False)
    assert len(res.tasks) == 4
    assert res.tasks[0].criteria['from'] == '20191001'
    assert res.tasks[1].criteria['from'] == '20191101'


def test_scrapper_deny_actions():
    res = Scrapper.run(MyScrapper(ScrapperData('place', {'place_id': '123123'}),
                                  deny_actions=[
                                      'tests.starling.test_app.MyAction2',
                                      'tests.starling.test_app.MyAction3'
                                  ]), json_output=False)
    assert len(res.tasks) == 2


def test_scrapper_json_output():
    res = json.loads(Scrapper.run(MyScrapper(ScrapperData('place', {'place_id': '123123'})), json_output=True))
    assert len(res['tasks']) == 4


def test_scrapper_with_attribute_error():
    with pytest.raises(AttributeError) as e:
        Scrapper.run(MyAttributeErrorScrapper(ScrapperData('place', {'place_id': '123123'})))

    assert "module" in str(e.value)



class ScrapingError(Exception):
    def __init__(self, message: str, extra: dict = None):
        self.message = message
        self.extra = extra

    def __str__(self):
        return f'{self.message} with {str(self.extra)}'


class AuthenticationError(ScrapingError):
    """ 인증 프로세스 실패 """


class RetryTaskExitError(ScrapingError):
    """
    Error in retry logic
    Use if no further progress is required
    """
    pass


class RetryTaskSkipAuthError(ScrapingError):
    """
    Error in retry logic
    Use if required except for authentication
    """
    pass


class RetryTaskError(ScrapingError):
    """
    Error in retry logic
    Use if required from authentication
    """
    pass



import abc

from starling.types import ActionData, TaskData, ScrapperData


class BlueprintAction(abc.ABC):
    def __init__(self, scrapper_data: ScrapperData, task_data: TaskData):
        self.scrapper_data: 'ScrapperData' = scrapper_data
        self.task_data: 'TaskData' = task_data
        self.action_data: 'ActionData' = ActionData()

    @abc.abstractmethod
    def fetch(self):
        pass

    @abc.abstractmethod
    def transform(self):
        pass

    def interval(self) -> int:
        return 0



# python3
# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides generic configuration handling.

It reads a yaml file from disk or yaml string, populates the yaml with the given
jinja variables, and loads the yaml variables into memory.
"""
import copy
import os
from typing import Text

import jinja2
import yaml


class _Config(object):
    """Config class to read configurations."""

    def __init__(self, file_name):
        self._items = None
        file_path = os.path.join(os.path.dirname(__file__), file_name)
        if not os.path.isfile(file_path):
            raise IOError(f'File not found: {file_path}')

        with open(file_path, 'r') as f:
            contents = ''.join(f.readlines())
            self._init(contents)

    def _init(self, yaml_string, **jinja_variables):
        """Populates the config object with variables defined in yaml_string.

        It renders jinja_variables into a YAML string and then loads the yaml into
        _items

        Args:
          yaml_string: yaml text to load.
          **jinja_variables: jinja variables to render inside of the yaml text.

        Raises:
          ConfigAlreadyInitializedError: if global config is already initialized.
              You must call `tear_down` before being able to `init` again.
          jinja2.exceptions.UndefinedError: if the yaml text requires jinja
              variables that are not included in jinja_variables.
          yaml.scanner.ScannerError: if the yaml text is formatted incorrectly.
        """
        if self._items is not None:
            raise ConfigAlreadyInitializedError()

        # Load the template
        template = jinja2.Environment(
            loader=jinja2.BaseLoader,
            undefined=jinja2.StrictUndefined).from_string(yaml_string)

        # Create the hash from the template and variables (set to {} if empty yaml)
        yaml_text = template.render(jinja_variables)
        self._items = yaml.safe_load(yaml_text) or {}

    def has_key(self, *keys: Text):
        """Returns a boolean indicating whether the key is defined in the config."""
        try:
            self.get(*keys)
            return True
        except KeyError:
            return False

    def get(self, *keys: Text):
        """Gets the value for given configuration key.

        If there is more than one key passed, each key will be searched
        as a nested property inside the parent key.

        Warning:
          `get` should not be called in any thread until `init` has completed.

        For example:
          Given a YAML string with content:
          PARENT:
            CHILD: val

          `get('PARENT', 'CHILD')` would return `val`.
          `get('PARENT')` would return the dictionary {'CHILD': 'val'}
          `get()` would return the full dictionary {'PARENT', {'CHILD': 'val'}}

        Args:
          *keys: keys tuple

        Returns:
          Configuration value corresponding to keys.

        Raises:
          KeyError: if key does not exists.
          ConfigNotInitializedError: if the global config hasn't been
              initialized yet.
        """
        if self._items is None:
            raise ConfigNotInitializedError()

        value = self._items
        for key in keys:
            if not isinstance(value, dict) or key not in value:
                raise KeyError(f'{key} is not found in config.')

            value = value.get(key)

        return copy.deepcopy(value)

    def _tear_down(self):
        """Cleanup for config object.

        Used for testing.

        This is useful if you want to initialize config multiple times.
        """
        self._items = None


class ConfigNotInitializedError(Exception):
    pass


class ConfigAlreadyInitializedError(Exception):
    pass


# Global Config instance
CONFIG = _Config(f'config.starling.{os.getenv("SCRAPPER_PROFILE", "development")}.yaml')



from setuptools import setup, find_packages

setup(
    name='kcd-starling',
    version='0.1.3',
    description='Skeleton for Scrapper',
    author='Terry Cho',
    author_email='terry@kcd.co.kr',
    packages=find_packages(exclude=['test']),
    zip_safe=False,
    include_package_data=True
)